#ifndef LZW_H
#define LZW_H

#include "lzwdico.h"	// dictionnaire spcifique de l'algorythme LZW
#include "crc32.h"	// classe de calcul d'un crc sur 32 bits
#include <stacks.h>	// classe template Borland pour une pile

//////////////////////////////////////////////////////////////////////////
// Classe ralisant la compression ou dcompression selon l'algorythme LZW.
// Les deux fonctions essentielles sont donc :
// 	compresse( const char*, const char* )
// 	decompresse( const char *, const char * )
// o les arguments sont videmment le nom des fichiers source et cible.
//
// La compression LZW est une compression exacte, en ce sens que la dcom-
// presssion du rsultat redonne un fichier absolument conforme  l'origi-
// nal (contrairement aux compressions d'images ou de sons).
//
// Une compression exacte peut faire appel  2 grandes ides :
// - chercher les redondances dans le fichier source, pour ne coder expli-
//	  citement que la premire occurrence, et se rfrer ensuite  celle-ci
//   par des "pointeurs" plus courts : rfrence  une entre donne par sa
//   position en amont, ou par une entre numrote dans un dictionnaire...
// - raccourcir au maximum la codification de chaque caractre, en adoptant
//   des codes courts pour les caractres frquents dans le source (au con-
//	  traire des codifications ASCII ou ANSI qui sont gnralistes, et codent
//   tous les caractres avec le mme nombre de bits).
//
// Rien n'empche de combiner les deux mthodes ; LZW est essentielement une
// mthode du premier type, mais elle utilise en partie aussi la deuxime i-
// de (notamment sous sa forme implmente ici : longueur de code variable).
//
// En voici les grandes lignes, par approches successives des dtails :
//
// - on cherche les redondances de chanes (parties communes) pour n'mettre
//   qu'une fois explicitement les radicaux frquents ; les autres occurren-
//   ces sont dsignes par leur numro d'entre dans un dictionnaire ;
//
// - le dictionnaire n'est pas crit avec le fichier cible ; l'algorythme de
//   constrution du dictionnaire est tel que le dcompresseur,  la lecture
//   du fichier comprim, sera capable de le reconstruire de faon identique
//    ce qu'il tait  chaque tape lors de la compression ;
//
// - au dpart, le dictionnaire contient 256 chanes de 1 caractre conte-
//   nant les caractres ASCII ; remarquer qu'ainsi le numro d'entre d'un
//   caractre dans le dictionnaire est gal  son code ASCII ;
//
// - on dcide d'mettre au dpart des codes de 9 bits au moins, ce qui cor-
//   respond  512 places dans le ditionnaire : les 256 caractres de base,
//   plus quelques codes spciaux qui seront des ordres pour le dcompresseur
//   (expliqus plus loin), plus de la place pour environ 250 chanes ;
//
// -  force d'entrer de nouvelles chanes dans le dictionnaire, il vient un
//   moment o 9 bits ne suffisent plus ; on dcide alors d'mettre des codes
//   de 10 bits ; il faut simplement le signaler au dcompresseur, et  cette
//   fin on met l'un des codes spciaux voqus plus haut ;
//
// -  chaque fois que la longueur du code est sature, on procde de mme ;
//
// - comme le dictionnaire est gr en mmoire, on ne peut l'augmenter ind-
//   finiment ; il y a donc une longueur maximale de code (comme 13 bits) au
//   del de laquelle le dictionnaire est rellement plein ; plusieurs tac-
//   tiques sont alors possibles, celle adopte ici tant de vider le dic-
//   tionnaire, en le signalant au dcompresseur par un autre code spcial ;
//
// - le dictionnaire se retrouve donc dans son tat de dpart : longueur de
//   de code  9 bits, 256 entres ASCII, plus quelques codes spciaux ;
//
// - il se remplit  nouveau ; on le revide si besoin est ; etc...
//
// - un troisime code spcial est rserv pour coder la fin de fichier.
//
// Ce qui est fascinant, c'est que finalement le compresseur peut faire ce
// qu'il veut, pourvu qu'il mette des codes spciaux pour en tenir inform
// le dcompreseur, qui fera exactement la mme chose, au mme moment.
//
// Le dictionnaire lui-mme est astucieux, en ce sens qu'il a la structure
// minimale ncessaire, afin de gagner du temps et de la place en mmoire.
// Il est dcrit dans lzwdico.h.
//
// Une pile est ncessaire, lors d'un cas particulier rencontr dans l'al-
// gorythme de dcompression ; elle est implmente  l'aide de la version
// template de la librairie de containers Borland en version 3.1.
//
// Enfin, on calcule un CRC du fichier source qui est stock en tte du fi-
// chier cible, avec d'autres donnes (longueur de l'original et du fichier
// compress, longueur minimale et maximale du code, en bits).
// Ceci permettra au dcompresseur de donner l'alerte si quoi que ce soit
// est anormal ; en particulier, on recalcule le crc aprs dcompression,
// et on le compare avec celui qui fut stock par le compresseur. En effet,
// il faut noter que la suppression de redondances, qui est l'essence mme
// de la compression, rend le message (le fichier compress) vulnrable 
// la moindre erreur ! ( tel point que les modems, aprs compression du
// message, lui ajoutent un peu de redondance sous forme de codes de cor-
// rection des erreurs ; c'est une situation presque absurde qui me semble
// appeler une rflexion sur les mthodes de codification ds le niveau du
// systme d'exploitation lui-mme...).
//
// D'autres dtails sur la ralisation de l'algorythme seront trouvs dans
// lzw.cpp. Quant  son principe, il est expos dans "La compression de don-
// nes", de Mark Nelson, diteur Dunod : un livre fondamental.
// J'y ai tout appris ; ma seule contribution est d'avoir transpos la pro-
// grammation en C classique (que je dteste) en encapsulant les choses dans
// des classes C++ (que j'aime), y compris pour les fonctionnalits d'entres
// et sorties de bits sur disque, dont nous utiliserons les aspects suivants :
// - ifBits, constructeur d'une classe flux d'entres de bits depuis disque ;
// - ofBits,      "          "     "    "   de sorties "   "  vers le disque ;
// - ifBits::readBits( unsigned int &, nbits ) lit nbits depuis le disque en
//   les plaant dans l'entier en argument 1 ;
// - ofBits::writeBits( unsigned int, nbits ) crit sur disque nbits trouvs
//   dans l'argument 1.
//
// Pour finir, il me faut reprendre un avertissement donn par Mark Nelson
// dans son livre : certaines parties de cet algorythme (notamment celles
// qualifies plus haut "d'astucieuses") sont couvertes par des brevets de
// la trs puissante compagnie Unisys et du troisime larron LZW,  savoir
// un certain Monsieur Terry Welsh. Comme quoi tout ce que nous faisons l
// ne devrait avoir d'autre but que pdagogique, ou  la rigueur un usage
// strictement personnel... Voyez Nelson pages 247 et 248.
//
//////////////////////////////////////////////////////////////////////////


#define 	Pile	BI_StackAsVector<unsigned char>


/////////
class LZW {
/////////
public :
				// Type numr dcrivant le rsultat des oprations
				enum				erreur	{ 	ras,			// pas d'erreur
													taille,		// avertissement taille compresse > original
													version,		// nombre de bits diffrent
													lgfich,     // longueur dcompresse diffrente
													crcfich,    // erreur crc sur fichier original
													crctete,    // erreur crc sur l'en-tte
													fsource,    // erreur ouverture source
													fcible		// erreur ouverture cible
												} ;
				const char *	message( erreur ) const ;	// message en clair

									// Constructeur et destructeur publics
									LZW	( unsigned int nombre_bits_final ) ;
								  ~LZW	() ;

				erreur			compresse	( const char * source, const char * cible ) ;
				erreur			decompresse	( const char * source, const char * cible  ) ;

#ifdef DEPANNAGE
				void				lister		() const ;
#endif

private :
				unsigned int	NBITS_FINAL ;	// longueur maximale du code
				LZWDico &		DICO ;         // dictionnaire, cf lzwdico.h
				Pile &			PILE ;			// pile de stockage de caractres

				void				empilerChaine	( LZWEntree::code ) ;	// cf le .cpp

				// Type (imbriqu priv) dcrivant l'en-tte du fichier compress
				// Pour une scurit maximale, on calcule 2 crc : celui de l'en-
				// tte (c'est  dire tout ce qu'elle contient en dehors de ce
				// crc), et celui du fichier source.
typedef		struct			tete_infos           // type utilis ci-dessous
				{
				unsigned long	crc_infos ;				// recevra le crc de ce qui suit

				unsigned char	nbits_depart,  		// en principe 9
									nbits_limite ;       // par exemple 13
				unsigned long	lg_fich_originel,    // longueur mmorise
									crc_fich_originel,   // crc mmoris
									lg_fich_comprime ;	// en-tte compris
				} ;

				// Variable recevant l'en-tte du fichier compress ; c'est une
				// union, pour tre accde de deux faons : d'une part comme
				// une structure tete_infos (type dfini ci-dessus), et d'autre
				// part comme un simple tableau d'octets (pour calculer le crc)
				struct										// en-tte du fichier comprim
				{
				union                               // ( le tableau ucars sert
				 {                                  // au calcul facile du crc )
				 tete_infos		infos ;              // cf ci-dessus le typedef
				 unsigned char	ucars[ sizeof( tete_infos ) ] ;
				 } ;
				}					en_tete ;

				CRC32				CRC_ORIG ;  			// crc calcul du fichier originel
																// (cf crc32.h et crc32.cpp)
				unsigned long	LG_ORIG ;				// longueur constate du fichier originel

				// quelques fonctions prives purement techniques (cf le .cpp)
				erreur			lireEnTete			( class ifBits & ) ;
				erreur			verifEnTete			() const ;
				void				ecrireEnTeteDeb	( class ofBits & ) ;
				erreur			ecrireEnTeteFin	( class ofBits & ) ;
				CRC32::crc 		CRCEnTete			()	const ;

} ;


#endif
